import io
import unittest
import unittest.mock as mock

from extra_metrics.config import ExtraMetricsConfiguration
from extra_metrics.test.fake_mocks import FakeQueryInterface

from extra_metrics.scripts import (
    provision_supervisord_runtime,
    validate_current_fw_version, ValidationExceptionCannotParseFileWaveVersion,
    ValidationExceptionWrongFileWaveVersion
)


class TestConfiguration(unittest.TestCase):
    def setUp(self):
        self.cfg = ExtraMetricsConfiguration()
        self.config_text = '''
            [extra_metrics]
            fw_server_hostname = abc
            fw_server_api_key = def
            '''
        buf = io.StringIO(self.config_text)
        self.cfg.read_configuration(buf)

    def test_configuration_defaults(self):
        self.assertEqual("abc", self.cfg.get_fw_api_server())
        self.assertEqual("def", self.cfg.get_fw_api_key())

    def test_configuration_changes_are_written(self):
        self.cfg.set_fw_api_key('hello world')
        self.cfg.set_polling_delay_seconds(42)
        buf = io.StringIO()
        self.cfg.write_configuration(buf)
        value = buf.getvalue()
        self.assertTrue("hello world" in value)
        self.assertTrue("42" in value)
        self.assertEqual(self.cfg.get_polling_delay_seconds(), "42")


def get_unparsable_version(self):
    return None, None, None


def get_incorrect_version(self):
    return 13, 3, 0


def get_correct_version(self):
    return 14, 0, 0


class TestRuntimeChecks(unittest.TestCase):
    def setUp(self):
        self.fw_query = FakeQueryInterface()

    @mock.patch('extra_metrics.test.fake_mocks.FakeQueryInterface.get_current_fw_version_major_minor_patch', get_unparsable_version)
    def test_validation_fails_due_to_bad_parsing(self):
        with self.assertRaises(ValidationExceptionCannotParseFileWaveVersion):
            validate_current_fw_version(self.fw_query)

    @mock.patch('extra_metrics.test.fake_mocks.FakeQueryInterface.get_current_fw_version_major_minor_patch', get_correct_version)
    def test_validation_succeeds(self):
        validate_current_fw_version(self.fw_query)

    @mock.patch('extra_metrics.test.fake_mocks.FakeQueryInterface.get_current_fw_version_major_minor_patch', get_incorrect_version)
    def test_validation_finds_incorrect_version(self):
        with self.assertRaises(ValidationExceptionWrongFileWaveVersion):
            validate_current_fw_version(self.fw_query)

    def test_can_run_supervisord_provisioning(self):
        provision_supervisord_runtime()



import asyncio
import atexit
import json
import zmq
import zmq.asyncio

from extra_metrics.logs import logger


def get_zmq_broker_xsub_endpoint(cfg):
    return f'tcp://{cfg.get_fw_api_server()}:20005'


class ZMQConnector:
    @classmethod
    def connect_to_broker(cls):
        cls.__instance = cls()
        atexit.register(cls.disconnect_from_broker)

    @classmethod
    def disconnect_from_broker(cls):
        cls.__instance.disconnect()

    def __init__(self, cfg, callback):
        self.__ctx: zmq.asyncio.Context = None
        self.__socket: zmq.asyncio.Socket = None
        self.cfg = cfg
        self.callback = callback
        asyncio.create_task(self.start())

    async def start(self):
        self.__ctx = zmq.asyncio.Context()
        self.__socket = self.__ctx.socket(zmq.SUB)

        # in compat mode; there is NO subscribe keypair data
        key_pair = self.cfg.get_zmq_subscribe_keypair()
        if key_pair is not None:
            self.__socket.curve_serverkey = key_pair.strip()  # the first line is the public key

            # for the client socket, the curve key pair will be regenerated all the time
            # the important part for "authentication" is the server "public" key
            self.__socket.curve_publickey, self.__socket.curve_secretkey = zmq.curve_keypair()

        self.__socket.connect(get_zmq_broker_xsub_endpoint(self.cfg))

        # Subscribe to all topics
        self.__socket.subscribe('')

        logger.info("filewave server event subscriber has started")

        asyncio.create_task(self.forward_events())

    async def forward_events(self):
        # forward JSON events into the callback
        while self.__socket:
            logger.debug('waiting for events from FileWave...')

            msg = await self.__socket.recv_multipart()
            topic = msg[0].decode('utf-8')
            body = json.loads(msg[3]) if msg[2] == b'json' else msg[3]

            try:
                logger.debug(f'new notification: {topic}, {body}')
                self.callback(topic, body)
            except Exception:
                # I really don't care what the consumer did with this
                pass

    def disconnect(self):
        self.__socket.close()
        self.__socket = None



from extra_metrics.config import ExtraMetricsConfiguration, read_config_helper
from extra_metrics.fwrest import FWRestQuery
from extra_metrics.logs import logger, init_logging

import os
import click
import shutil
import pkg_resources
import sys


class ValidationExceptionCannotParseFileWaveVersion(Exception):
    pass


class ValidationExceptionWrongFileWaveVersion(Exception):
    pass


@click.group()
def cli():
    click.echo("FileWave Extra Metrics configuration.")


delay_30m = 60 * 30


@cli.command('validate', help="Validates the configuration assuming you are running this on the FileWave Server")
@click.option('-c', '--config-path', 'config_path', default=ExtraMetricsConfiguration.DEFAULT_CFG_FILE_LOCATION, help='the full path to the configuration file')
@click.option('-a', '--api-key', 'api_key', default=None, help='the FileWave API Key with appropriate rights to create groups/queries')
@click.option('-e', '--external-dns-name', 'external_dns_name', default=None, help='the externally visible DNS name for the filewave server')
@click.option('-i', '--interval', 'polling_interval', default=delay_30m, help='the seconds delay between successive queries against the FileWave system ()')
@click.option('-s', '--skip-local-validation', is_flag=True, default=False, help='skips the local server validation checks, useful for running off a fw server')
def install_into_environment(config_path, api_key, external_dns_name, polling_interval, skip_local_validation):
    init_logging()

    cfg = ExtraMetricsConfiguration()
    dirname = os.path.dirname(config_path)
    if not os.path.exists(dirname):
        logger.error(f"The directory for the configuration file does not exist: {dirname}")
        return

    if not os.path.exists(config_path) and os.path.isfile(config_path):
        if not os.access(config_path, os.W_OK):
            logger.error(f"The configuration file cannot be written to {config_path} - does this user have access?")
            return

    try:
        read_config_helper(cfg)
    except FileNotFoundError:
        if api_key is None or external_dns_name is None:
            logger.error("When there is no configuration file you must specify an API key and external DNS name, which will then be stored in the config file")
            return

    assert cfg.section is not None

    if api_key is not None:
        cfg.set_fw_api_key(api_key)
    if external_dns_name is not None:
        cfg.set_fw_api_server(external_dns_name)

    if polling_interval is not None:
        cfg.set_polling_delay_seconds(polling_interval)

    try:
        with open(config_path, 'w+') as f:
            cfg.write_configuration(f)
            logger.info(f"saved configuration to file: {config_path}")
    except Exception as e:
        logger.error("Unable to write the configuration file - normally this command requires sudo/root privs, did you use sudo?")
        logger.error(e)
        return

    if not skip_local_validation:
        try:
            provision_dashboards_into_grafana(cfg.get_fw_api_server())
            provision_prometheus_scrape_configuration()
            provision_supervisord_runtime()
        except Exception as e:
            logger.error("Error during provisioning of prometheus/grafana, are you using sudo?")
            logger.error(e)
            return

    q = FWRestQuery(cfg.get_fw_api_server(), cfg.get_fw_api_key())
    major, minor, patch = validate_runtime_requirements(q)

    log_config_summary(cfg, major, minor, patch)


def log_config_summary(cfg, major, minor, patch):
    logger.info("")
    logger.info("Extra Metrics - Configuration Summary")
    logger.info("=====================================")
    logger.info(f"External DNS     : {cfg.get_fw_api_server()}")
    logger.info(f"API Key          : {cfg.get_fw_api_key()}")
    logger.info(f"FileWave Server  : {major}.{minor}.{patch}")
    logger.info(f"Polling Interval : {cfg.get_polling_delay_seconds()} sec")


def provision_supervisord_runtime():
    supervisord_dir = os.path.join("/usr/local/etc/filewave/supervisor/", "extras")
    if not os.path.exists(supervisord_dir):
        os.makedirs(supervisord_dir)

    # lets go hunting, I want to find executable 'extra_metrics_run' - which is installed by setuptool,
    # but I don't know where the whole thing is being run from... so I'm looking for bin/extra_metrics_run
    # starting with the sys.executable directory
    exec_path = os.path.dirname(sys.executable)
    full_extra_metrics_run_path = "extra-metrics-run"
    for f in os.listdir(exec_path):
        if f == "extra-metrics-run":
            full_extra_metrics_run_path = os.path.join(exec_path, full_extra_metrics_run_path)

    data = pkg_resources.resource_string("extra_metrics.cfg", "extra_metrics_supervisord.conf").decode('utf-8')
    provisioning_file = os.path.join(supervisord_dir, 'extra_metrics_supervisord.conf')
    with open(provisioning_file, "w+") as f:
        new_data = data.replace(r'${EXTRA_METRICS_RUN}', full_extra_metrics_run_path)
        f.write(new_data)


def provision_dashboards_into_grafana(fw_server_dns_name):
    # if the expected dashboards DO NOT exist in the right path, moan about this and go ahead and copy them..
    grafana_dashboard_deployment_dir = os.path.join("/usr/local/etc/filewave/grafana/provisioning", "dashboards")
    if not os.path.exists(grafana_dashboard_deployment_dir):
        logger.error(f"The Grafana dashboard deployment directory ({grafana_dashboard_deployment_dir}) does not exist; is this version 14+ of FileWave?")
        return

    # check each file is there... overwrite regardless (helps on upgrade I suppose)
    for dashboard_file in pkg_resources.resource_listdir("extra_metrics", "dashboards"):
        if dashboard_file.endswith(".json"):
            data = pkg_resources.resource_string("extra_metrics.dashboards", dashboard_file).decode('utf-8')
            provisioning_file = os.path.join(grafana_dashboard_deployment_dir, dashboard_file)
            with open(provisioning_file, 'w+') as f:
                # load up the dashboard and replace the ${VAR_SERVER} with our config value of the external DNS
                new_data = data.replace('${VAR_SERVER}', fw_server_dns_name)
                f.write(new_data)
                logger.info(f"wrote dashboard file: {provisioning_file}")


def provision_prometheus_scrape_configuration():
    prometheus_dir = os.path.join("/usr/local/etc/filewave/prometheus/conf.d/jobs", "http")
    if not os.path.exists(prometheus_dir):
        logger.error(f"The Prometheus directory ({prometheus_dir}) does not exist; is this version 14+ of FileWave?")
        return

    for yaml_file in pkg_resources.resource_listdir("extra_metrics", "cfg"):
        if yaml_file.endswith(".yml"):
            data = pkg_resources.resource_string("extra_metrics.cfg", yaml_file)
            provisioning_file = os.path.join(prometheus_dir, yaml_file)
            with open(provisioning_file, 'wb') as f:
                f.write(data)
            shutil.chown(provisioning_file, user="apache", group="apache")


def validate_current_fw_version(fw_query):
    major, minor, patch = fw_query.get_current_fw_version_major_minor_patch()
    if major is None:
        raise ValidationExceptionCannotParseFileWaveVersion("Failed to detection which version of FileWave is running")

    if major < 14:
        err = f"You must be running FileWave version 14 or above - I detected a FileWave server version: {major}.{minor}.{patch}"
        logger.error(err)
        raise ValidationExceptionWrongFileWaveVersion(err)

    return major, minor, patch


def validate_current_api_keys_rights(fw_query):
    pass


def validate_runtime_requirements(q):
    # TODO: can I upgrade the grafana pie chart automatically?
    validate_current_api_keys_rights(q)
    return validate_current_fw_version(q)




import json
import pandas as pd

from extra_metrics.test.test_queries import \
    MAIN_GROUP_ID, MAIN_GROUP_NAME, MAIN_GROUP_PARENT_ID

from extra_metrics.test.test_queries import \
    test_query_app_adobe_acrobat_reader_win, \
    test_query_app_wrong_group_zoom_win, \
    test_query_app_zoom_win, \
    test_win_app_generic_should_not_load

pd.set_option('display.precision', 3)
pd.set_option('display.expand_frame_repr', False)


class FakeRequest:
    def __init__(self, string_data):
        self.data = string_data

    def json(self):
        return json.loads(self.data)


class FakeQueryInterface:
    TEST_QUERY_APPS = 1
    TEST_QUERY_DATA_INVALID = 99
    TEST_QUERY_DATA_ALMOST_VALID1 = 98
    TEST_QUERY_DATA_ALMOST_VALID2 = 97

    def __init__(self, create_inventory_callback=None):
        self.create_inventory_callback = create_inventory_callback
        super()

    def get_current_fw_version_major_minor_patch(self):
        raise Exception("not implemented - please mock this")

    def create_inventory_query(self, json_obj):
        if self.create_inventory_callback:  # pragma: no branch
            self.create_inventory_callback(json_obj)

    def get_software_updates_web_ui_j(self):
        t = '''{
            "results": [
            ]
        }'''
        return json.loads(t)

    def ensure_inventory_query_group_exists(self, name_of_query):
        assert name_of_query is not None
        my_data = '''{
            "name": "%s",
            "parent": %d,
            "id": %d
        }''' % (MAIN_GROUP_NAME, MAIN_GROUP_PARENT_ID, MAIN_GROUP_ID)
        return json.loads(my_data), False

    def get_all_inventory_queries(self):
        # returns some good, some bad... some in our group, some not.
        text_queries = [test_query_app_adobe_acrobat_reader_win,
                        test_win_app_generic_should_not_load,
                        test_query_app_wrong_group_zoom_win,
                        test_query_app_zoom_win]

        return [json.loads(t) for t in text_queries]

    def get_definition_for_query_id_j(self, q_id):
        if q_id == FakeQueryInterface.TEST_QUERY_DATA_INVALID:
            return None

        if q_id == FakeQueryInterface.TEST_QUERY_DATA_ALMOST_VALID1:
            badly_formed_query = '''{
                "favorite": true,
                "fields": [
                    {
                        "column": "filewave_id",
                        "component": "Client"
                    },
                    {
                        "column": "filewave_client_name",
                        "component": "Client"
                    },
                    {
                        "column": "version",
                        "component": "Update"
                    }
                ],
                "main_component": "Update"
                }'''
            return json.loads(badly_formed_query)

        if q_id == FakeQueryInterface.TEST_QUERY_DATA_ALMOST_VALID2:
            badly_formed_query = '''{
                "favorite": true,
                "fields": [
                    {
                        "column": "filewave_id",
                        "component": "Client"
                    },
                    {
                        "column": "filewave_client_name",
                        "component": "Client"
                    },
                    {
                        "column": "version",
                        "component": "Update"
                    }
                ],
                "name": "this has a name",
                "main_component": "Update"
                }'''
            return json.loads(badly_formed_query)

        if q_id == 66:
            badly_formed_query = '''{
                "favorite": true,
                "fields": [
                    {
                        "column": "filewave_id",
                        "component": "Client"
                    },
                    {
                        "column": "filewave_client_name",
                        "component": "Client"
                    },
                    {
                        "column": "version",
                        "component": "Update"
                    }
                ],
                "main_component": "Update",
                "id": 66,
                "name": "extra metrics - software patch"
                }'''
            return json.loads(badly_formed_query)

        if q_id == 55:
            well_formed_query = '''{
                "favorite": true,
                "fields": [
                    {
                        "column": "device_id",
                        "component": "Client"
                    },
                    {
                        "column": "name",
                        "component": "Application"
                    },
                    {
                        "column": "version",
                        "component": "Application"
                    }
                ],
                "main_component": "Update",
                "name": "extra metrics - software patch",
                "id": 55
            }'''
            return json.loads(well_formed_query)

        all_queries = self.get_all_inventory_queries()
        for a_query in all_queries:
            if a_query["id"] == q_id:
                return a_query
        return None

    def get_results_for_query_id(self, query_id):
        if query_id == 105:
            return FakeRequest('''{
                "offset": 0,
                "fields": [
                    "Application_product_id",
                    "Application_name",
                    "Application_version",
                    "Application_size",
                    "Application_install_size",
                    "Application_is_validated",
                    "Client_filewave_client_name",
                    "Client_device_id"
                ],
                "values": [
                    [
                        "{abcdefg}",
                        "Zoom",
                        "12.1",
                        null,
                        6459167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95adf8b12bb24e4f7139"
                    ],
                    [
                        "{AC76BA86-7AD7-1033-7B44-AC0F074E4100}",
                        "Zoom",
                        "20",
                        null,
                        339167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95adf8b12bb24e4f7139"
                    ],
                    [
                        "{AC76BA86-7AD7-1033-7B44-AC0F074E4100}",
                        "Zoom App",
                        "20.1",
                        null,
                        339167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95bdf8b12bb24e4f7139"
                    ]

                ],
                "filter_results": 3,
                "total_results": 3,
                "version": 1
            }''')

        if query_id == FakeQueryInterface.TEST_QUERY_APPS or query_id == 101:
            return FakeRequest('''{
                "offset": 0,
                "fields": [
                    "Application_product_id",
                    "Application_name",
                    "Application_version",
                    "Application_size",
                    "Application_install_size",
                    "Application_is_validated",
                    "Client_filewave_client_name",
                    "Client_device_id"
                ],
                "values": [
                    [
                        "{abcdefg}",
                        "Adobe Acrobat Reader DC",
                        "12.1",
                        null,
                        6459167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95adf8b12bb24e4f7139"
                    ],
                    [
                        "{AC76BA86-7AD7-1033-7B44-AC0F074E4100}",
                        "Adobe Acrobat Reader DC",
                        "20.009.20067",
                        null,
                        339167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95adf8b12bb24e4f7139"
                    ],
                    [
                        "{AC76BA86-7AD7-1033-7B44-AC0F074E4100}",
                        "Adobe Acrobat Reader DC",
                        "20.009.20067",
                        null,
                        339167232,
                        null,
                        "hades2",
                        "4d9fe4590232f39b783d95bdf8b12bb24e4f7139"
                    ]

                ],
                "filter_results": 3,
                "total_results": 3,
                "version": 1
            }''')



from prometheus_client import start_http_server
from extra_metrics.logs import logger, init_logging
from extra_metrics.scripts import log_config_summary
from periodic import Periodic
import pandas as pd
import asyncio
import json

from extra_metrics.application import ApplicationQueryManager
from extra_metrics.softwarepatches import SoftwarePatchStatus
from extra_metrics.devices import PerDeviceStatus
from extra_metrics.fwrest import FWRestQuery
from extra_metrics.fw_zmq_eventsub import ZMQConnector
from extra_metrics.config import ExtraMetricsConfiguration, read_config_helper

# TODO: link to devices affected in the dashboard is wrong; we can fix that!  https://${server}/reports/46/details/

# DONE: can I subscribe to FW query updates to run the app queries again?  or just the single app query?

# TODO: if the configuration of the FW server changes, how does this product keep up?

# TODO: mock the fwrestquery API itself, to prove that all methods produce the right reactions.

# TODO: DEVICE HEALTH show the reasons/state of health of a device in (my app or) custom fields so it can be reported on

# TODO: consider alerts for devices entering a non-healthy state for the first time today

# TODO: consider how I might achieve a cannibalization test for devices?

# TODO: write documentation on arch and reasoning... perhaps using the Wiki in GitHub

# TODO: languages / translation?

pd.set_option('display.precision', 3)
pd.set_option('display.expand_frame_repr', False)


class MainRuntime:
    def __init__(self, logger):
        self.cfg = None
        self.zmq_sub = None
        self.app_qm = None
        self.software_patches = None
        self.per_device = None
        self.logger = logger
        self.rerun_data_collection = True

    def init_services(self):
        self.cfg = ExtraMetricsConfiguration()
        read_config_helper(self.cfg)

        self.fw_query = FWRestQuery(
            hostname=self.cfg.get_fw_api_server(),
            api_key=self.cfg.get_fw_api_key())

        self.app_qm = ApplicationQueryManager(self.fw_query)
        self.software_patches = SoftwarePatchStatus(self.fw_query)
        self.per_device = PerDeviceStatus(self.fw_query)

        self.zmq_sub = ZMQConnector(self.cfg, lambda topic, payload: self.event_callback(topic, payload))

    def event_callback(self, topic, payload):
        self.logger.info(f"event received: {topic}")

        interesting_topics = [
            "/server/update_model_finished",
            "/api/auditlog",
        ]

        debug_topics = interesting_topics + [
            "/inventory/inventory_query_changed",
            "/server/change_packets"
        ]

        if topic in debug_topics:
            pretty_print_json = json.dumps(payload, indent=4)
            logger.info(f"topic: {topic}")
            logger.info(f"payload: {pretty_print_json}")

        if topic in interesting_topics:
            if topic == "/api/auditlog" and "Report Created" not in payload["message"]:
                return

            # set a flag in state indicating that the data collection should run imminently
            logger.info(f"topic {topic} fired; will re-queue data collection")
            self.rerun_data_collection = True

    async def validate_and_collect_data(self):
        self.app_qm.validate_query_definitions()
        self.app_qm.collect_application_query_results()

        self.software_patches.collect_patch_data_status()
        self.software_patches.collect_patch_data_per_device()

        self.per_device.collect_client_data(self.software_patches)

        self.rerun_data_collection = False


async def create_program_and_run_it():
    init_logging()

    prog = MainRuntime(logger)
    prog.init_services()
    start_http_server(8000)

    host = prog.cfg.get_fw_api_server()
    poll_interval = prog.cfg.get_polling_delay_seconds()
    logger.info(f"Extra Metrics - connecting to {host}, using poll interval of {poll_interval} sec")

    major, minor, patch = prog.fw_query.get_current_fw_version_major_minor_patch()
    log_config_summary(prog.cfg, major, minor, patch)

    if major is None:
        logger.error("Unable to reach FileWave server, aborting...")
        return

    p = Periodic(poll_interval, prog.validate_and_collect_data)
    await p.start()

    while(True):
        if prog.rerun_data_collection:
            await prog.validate_and_collect_data()
        await asyncio.sleep(1)


def serve_and_process():
    asyncio.run(create_program_and_run_it())


if __name__ == "__main__":
    serve_and_process()



import unittest
from extra_metrics.fwrest import FWRestQuery
from extra_metrics.config import read_config_helper, ExtraMetricsConfiguration


class FWQueryTestCase(unittest.TestCase):
    def test_query_init(self):
        fq = FWRestQuery("a", "b")
        self.assertEqual(fq.hostname, "a")
        self.assertEqual(fq.api_key, "b")

    def test_query_inventory_stub(self):
        fq = FWRestQuery("a", "b")
        self.assertEqual("https://a:20445/inv/api/v1/def",
                         fq._fw_run_inv_query("def"))
        self.assertEqual("https://a/api/xyz", fq._fw_run_web_query("xyz"))

    def test_live_get_software_patches_j(self):
        cfg = ExtraMetricsConfiguration()
        read_config_helper(cfg)

        fw_q = FWRestQuery(cfg.get_fw_api_server(), cfg.get_fw_api_key())

        r = fw_q.get_software_patches_j()
        print(r)



import unittest
from prometheus_client import REGISTRY

from extra_metrics.test.fake_mocks import FakeQueryInterface
from extra_metrics.test.test_queries import MAIN_GROUP_ID
from extra_metrics.application import (
    ApplicationQueryManager,
    ApplicationUsageRollup
)


class ExtraMetricsTestCase(unittest.TestCase):
    def setUp(self):
        self.fw_query = FakeQueryInterface()

    def test_app_manager_rolls_queries_and_populates_metrics(self):
        app_mgr = ApplicationQueryManager(self.fw_query)
        app_mgr.validate_query_definitions()
        app_mgr.collect_application_query_results()

        after = REGISTRY.get_sample_value('extra_metrics_application_version',
                                          labels={"query_name": "Adobe Acrobat Reader Win",
                                                  "application_version": "12.1", "query_id": "101"})
        self.assertEqual(1, after)

        after = REGISTRY.get_sample_value('extra_metrics_application_version',
                                          labels={"query_name": "Adobe Acrobat Reader Win",
                                                  "application_version": "20.009.20067", "query_id": "101"})
        self.assertEqual(2, after)

        after = REGISTRY.get_sample_value('extra_metrics_application_version',
                                          labels={"query_name": "Zoom Win", "application_version": "20.1",
                                                  "query_id": "105"})
        self.assertEqual(1, after)

    def test_app_manager_query_validation(self):
        app_mgr = ApplicationQueryManager(self.fw_query)
        self.assertEqual(len(app_mgr.app_queries), 0)
        app_mgr.validate_query_definitions()
        self.assertEqual(len(app_mgr.app_queries), 2)

    def test_app_manager_invalid_query_cannot_validate(self):
        app_mgr = ApplicationQueryManager(self.fw_query)
        self.assertFalse(app_mgr.is_query_valid(
            FakeQueryInterface.TEST_QUERY_DATA_INVALID))
        self.assertFalse(app_mgr.is_query_valid(
            FakeQueryInterface.TEST_QUERY_DATA_ALMOST_VALID1))
        self.assertFalse(app_mgr.is_query_valid(
            FakeQueryInterface.TEST_QUERY_DATA_ALMOST_VALID2))

    def test_app_manager_doesnt_load_crap_queries(self):
        app_mgr = ApplicationQueryManager(self.fw_query)
        app_mgr.retrieve_all_queries_in_group(MAIN_GROUP_ID)
        self.assertEqual(len(app_mgr.app_queries), 2)
        self.assertTrue(120 not in app_mgr.app_queries.keys())
        self.assertTrue(101 in app_mgr.app_queries.keys())
        self.assertTrue(105 in app_mgr.app_queries.keys())

    def test_app_mgr_accepts_valid_inventory_query(self):
        app_mgr = ApplicationQueryManager(self.fw_query)
        result = app_mgr.is_query_valid(55)
        self.assertTrue(result)

    def test_app_mgr_rejects_invalid_inventory_query(self):
        app_mgr = ApplicationQueryManager(self.fw_query)
        result = app_mgr.is_query_valid(66)
        self.assertFalse(result)

    def test_app_rollup(self):
        thing = ApplicationUsageRollup(FakeQueryInterface.TEST_QUERY_APPS, [
            "Application_name", "Application_version"], "Client_device_id")
        thing.exec(self.fw_query)

        # examine the rollup results to prove this worked
        r = thing.results()
        self.assertEqual(len(r), 2)

        item = r[0]
        item_name = item[0]
        item_version = item[1]
        item_count = item[2]

        self.assertEqual(item_name, "Adobe Acrobat Reader DC")
        self.assertEqual(item_version, "12.1")
        self.assertEqual(item_count, 1)

        item = r[1]
        item_name = item[0]
        item_version = item[1]
        item_count = item[2]

        self.assertEqual(item_name, "Adobe Acrobat Reader DC")
        self.assertEqual(item_version, "20.009.20067")
        self.assertEqual(item_count, 2)

    def test_app_mgr_can_create_queries(self):
        loaded_data = []

        def test_query_load(json_data):
            nonlocal loaded_data
            loaded_data.append(json_data)

        my_query = FakeQueryInterface(test_query_load)
        app_mgr = ApplicationQueryManager(my_query)
        app_mgr.create_default_queries_in_group(MAIN_GROUP_ID)

        self.assertEqual(9, len(loaded_data))



from prometheus_client import Gauge
import pandas as pd
from extra_metrics.logs import logger

software_updates_by_state = Gauge('extra_metrics_software_updates_by_state',
    'buckets of all the software updates, according to the number of devices in each state',
    ["state"])

software_updates_by_device = Gauge('extra_metrics_software_updates_by_device',
    'list of devices and the number of [critical] updates they need to have installed',
    ["device_name", "device_id", "is_update_critical"])

software_updates_by_update = Gauge('extra_metrics_software_updates_by_update',
    'list of updates and the number of devices asking to have installed',
    ["update_name", "fw_id"])

software_updates_by_platform = Gauge('extra_metrics_software_updates_by_platform',
    'list of platforms and the number of [critical] updates they have available',
    ["platform_name", "is_update_critical"])


class PerDevicePatchState:
    def __init__(self, client_id):
        self.client_id = client_id
        self.client_name = None
        self.critical_patch_count = 0
        self.standard_patch_count = 0


class SoftwarePatchStatus:
    def __init__(self, fw_query):
        self.fw_query = fw_query
        self.state_by_patch = {}
        # a dictionary of the PerDevicePatchState
        self.state_by_device = {}

    def get_perdevicestate_for_client_id(self, client_id):
        if client_id not in self.state_by_device:
            self.state_by_device[client_id] = PerDevicePatchState(client_id)
        return self.state_by_device.get(client_id, None)

    def collect_patch_data_status(self):
        self.state_by_patch = {}

        j = self.fw_query.get_software_updates_web_ui_j()
        if j is None:
            return

        if "results" not in j or len(j["results"]) == 0:
            logger.info("no results for software update patch status received from FileWave server")
            return None

        columns = [
            "update_name",
            "fw_id",
            "platform",
            "unassigned",
            "assigned",
            "completed",
            "remaining",
            "warning",
            "error"
        ]

        values = [
        ]

        res = j["results"]
        for item in res:
            update_id = item['update_id']
            self.state_by_patch[update_id] = item

            acc = item["assigned_clients_count"]

            item_name = item["name"]
            update_name = item_name["display_value"]
            if item_name["os_type"] == "OSX":
                update_name += f" ({update_id})"

            values.append([
                update_name,
                item["id"],
                item["platform"],
                item["count_unassigned"],
                acc["assigned"],
                acc["completed"],
                acc["remaining"],
                acc["warning"],
                acc["error"]
            ])

        df = pd.DataFrame(values, columns=columns)

        per_update_totals = df.groupby(["update_name", "fw_id"], as_index=False)["unassigned"].sum()
        for item in per_update_totals.to_numpy():
            software_updates_by_update.labels(item[0], item[1]).set(item[2])

        totals = df.sum(0, numeric_only=True)

        # goal: number of software updates, grouped by status:
        # status values:
        # NEW: count_unassigned
        # IN PROGRESS: count_assigned
        # FAILED: assigned_clients.error + assigned_clients.warning
        # DONE: assigned_clients.remaining == 0

        software_updates_by_state.labels('Unassigned').set(totals["unassigned"])
        software_updates_by_state.labels('Assigned').set(totals["assigned"])
        software_updates_by_state.labels('Completed').set(totals["completed"])
        software_updates_by_state.labels('Remaining').set(totals["remaining"])
        software_updates_by_state.labels('Warning').set(totals["warning"])
        software_updates_by_state.labels('Error').set(totals["error"])

    def collect_patch_data_per_device(self):
        self.state_by_device = {}

        j = self.fw_query.get_software_patches_j()
        if j is None:
            return

        if "values" not in j or len(j["values"]) == 0:
            logger.info("no results for software update patch status per device received from FileWave server")
            return None

        df = pd.DataFrame(j["values"], columns=j["fields"])
        platform_mapping = {
            "0": "Apple",
            "1": "Microsoft"
        }

        ru = df.groupby(["Client_filewave_client_name", "Client_filewave_id", "Update_critical"], as_index=False)["Update_name"].count()
        for i in ru.to_numpy():
            client_name = i[0]
            client_id = int(i[1])
            is_crit = i[2]
            update_count = i[3]
            logger.info(f"patches, device: {client_name}/{client_id}, critical: {is_crit}, number of updates: {update_count}")
            software_updates_by_device.labels(client_name, client_id, is_crit).set(update_count)

            # update internal counter state.... we are tracking total number of standard/critical updates per client.
            obj = self.get_perdevicestate_for_client_id(client_id)
            obj.client_name = client_name
            if is_crit:
                obj.critical_patch_count += update_count
            else:
                obj.standard_patch_count += update_count

        ru = df.groupby(["Update_platform", "Update_critical"], as_index=False)["Update_name"].count()
        for i in ru.to_numpy():
            platform_name = platform_mapping[i[0]]
            is_crit = i[1]
            update_count = i[2]
            software_updates_by_platform.labels(platform_name, is_crit).set(update_count)



import unittest
from extra_metrics.compliance import ClientCompliance


class TestClientComplianceCase(unittest.TestCase):
    def test_compliance_with_no_data(self):
        c = ClientCompliance(None, None, None)
        self.assertEqual(c.get_checkin_compliance(),
                         ClientCompliance.STATE_UNKNOWN)
        self.assertEqual(c.get_disk_compliance(),
                         ClientCompliance.STATE_UNKNOWN)
        self.assertEqual(c.get_compliance_state(),
                         ClientCompliance.STATE_UNKNOWN)
        self.assertEqual(c.get_patch_compliance(),
                         ClientCompliance.STATE_UNKNOWN)

    def test_compliance_patch_status_ok(self):
        c = ClientCompliance(None, None, None, 0, 0)
        self.assertEqual(c.get_patch_compliance(),
                         ClientCompliance.STATE_OK)

    def test_compliance_patch_status_failure(self):
        # 1 critical package, no other packages
        c = ClientCompliance(None, None, None, 1, 0)
        self.assertEqual(c.get_patch_compliance(),
                         ClientCompliance.STATE_ERROR)
        # 0 critical packages, 1 other package
        c = ClientCompliance(None, None, None, 0, 1)
        self.assertEqual(c.get_patch_compliance(),
                         ClientCompliance.STATE_WARNING)

    def test_compliance_checkin_less_than_7_is_ok(self):
        c = ClientCompliance(None, None, 6)
        self.assertEqual(c.get_checkin_compliance(), ClientCompliance.STATE_OK)
        c.last_checkin_days = 7
        self.assertEqual(c.get_checkin_compliance(),
                         ClientCompliance.STATE_WARNING)

    def test_compliance_checkin_less_than_14_is_warning(self):
        c = ClientCompliance(None, None, 13)
        self.assertEqual(c.get_checkin_compliance(),
                         ClientCompliance.STATE_WARNING)
        c.last_checkin_days = 14
        self.assertEqual(c.get_checkin_compliance(),
                         ClientCompliance.STATE_ERROR)

    def test_compliance_disk_space_free_more_than_20pcnt(self):
        # < 20% left is warning
        # < 5% left is critical, or less than 5g
        c = ClientCompliance(100, 25, 3)
        self.assertEqual(c.get_disk_compliance(), ClientCompliance.STATE_OK)
        c = ClientCompliance(100, 20, 3)
        self.assertEqual(c.get_disk_compliance(), ClientCompliance.STATE_OK)

    def test_compliance_disk_space_free_between_5_and_20pcnt(self):
        c = ClientCompliance(100, 5, 3)
        self.assertEqual(c.get_disk_compliance(),
                         ClientCompliance.STATE_WARNING)
        c = ClientCompliance(100, 19, 3)
        self.assertEqual(c.get_disk_compliance(),
                         ClientCompliance.STATE_WARNING)

    def test_compliance_disk_space_free_less_than_5pcnt(self):
        c = ClientCompliance(100, 4.9, 3)
        self.assertEqual(c.get_disk_compliance(), ClientCompliance.STATE_ERROR)

    def test_compliance_value_is_max_of_either(self):
        # enforce that it's max() of either value
        c1 = ClientCompliance(100, 5, 3)  # WARNING for DISK
        c2 = ClientCompliance(100, 40, 500)  # ERROR for CHECKIN DAYS
        self.assertTrue(c2.get_checkin_compliance() > c1.get_disk_compliance())
        self.assertEqual(c2.get_compliance_state(),
                         ClientCompliance.STATE_ERROR)



import unittest
from extra_metrics.test.fake_mocks import FakeQueryInterface
from extra_metrics.softwarepatches import SoftwarePatchStatus


class TestSoftwarePatchFetching(unittest.TestCase):
    def setUp(self):
        self.fw_query = FakeQueryInterface()

    def test_that_zero_data_doesnt_cause_bad_things(self):
        mgr = SoftwarePatchStatus(self.fw_query)
        mgr.collect_patch_data_status()
